/* This file is part of SLB
   Copyright (C) 2011 Sergey Poznyakoff

   SLB is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   SLB is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with SLB.  If not, see <http://www.gnu.org/licenses/>. */

#include "slb.h"

static struct symtab *expr_table;

/* Register an expression. */
int
register_expression(struct slb_expression *expr)
{
	struct slb_expression *s;
	int install = 1;
	
	if (!expr_table) {
		expr_table =
			symtab_create_default(sizeof(struct slb_expression));
		if (!expr_table)
			xalloc_die();
	}

	s = symtab_lookup_or_install(expr_table, expr, &install);
	if (!s) {
		logmsg(LOG_CRIT, _("cannot register expression: %s"),
		       symtab_strerror(errno));
		exit(EX_SOFTWARE);
	}

	if (!install) {
		grecs_error (&expr->ex_locus, 0,
			     _("redefinition of expression `%s'"),
			     expr->ex_name);
		grecs_error (&s->ex_locus, 0,
			     _("previously defined here"));
		return 1;
	}

	memcpy((char*)s + offsetof(struct slb_expression, ex_text),
	       (char*)expr + offsetof(struct slb_expression, ex_text),
	       sizeof(struct slb_expression) -
	       offsetof(struct slb_expression, ex_text));

	return 0;
}

struct slb_expression *
expression_lookupz(const char *name)
{
	struct slb_expression ex, *s;
	
	if (!expr_table)
		return NULL;
	ex.ex_name = name;

	s = symtab_lookup_or_install(expr_table, &ex, NULL);
	if (!s) {
		if (errno == ENOENT)
			return NULL;
		else {
			logmsg(LOG_CRIT, _("expression lookup: %s"),
			       symtab_strerror(errno));
			exit(EX_SOFTWARE);
		}
	}
	return s;
}

struct slb_expression *
expression_lookup(const char *name, size_t len)
{
	struct slb_expression ex, *s;
	static char *namebuf;
	static size_t namelen;
	
	if (!expr_table)
		return NULL;
	if (len + 1 >= namelen) {
		namelen = len + 1;
		namebuf = xrealloc(namebuf, namelen);
	}
	memcpy(namebuf, name, len);
	namebuf[len] = 0;
	ex.ex_name = namebuf;

	s = symtab_lookup_or_install(expr_table, &ex, NULL);
	if (!s) {
		if (errno == ENOENT)
			return NULL;
		else {
			logmsg(LOG_CRIT, _("expression lookup: %s"),
			       symtab_strerror(errno));
			exit(EX_SOFTWARE);
		}
	}
	return s;
}

struct slb_expression *
copy_expression(struct slb_expression *src, struct symtab *vt)
{
	struct slb_expression *dst = xmalloc(sizeof(*dst));
	memcpy(dst, src, sizeof(*dst));
	if (vt)
		dst->ex_vartab = vt;
	else
		dst->ex_vartab = symtab_create_default(sizeof(struct slb_varref));
	dst->ex_tree = copy_node(src->ex_tree, dst->ex_vartab);
	return dst;
}


#define MAX(a,b) (((a) > (b)) ? (a) : (b))

static void
nodefixup(struct slb_node *node, struct symtab *vit,
	  struct slb_expression *expr)
{
	struct slb_varinstance *vi;
	struct slb_node *np;
	int min_eval = 0;
	
	switch (node->n_type) {
	case node_unknown:
		abort();
		
	case node_variable:
		vi = varinst_lookupz(vit, node->v.n_var->vr_sym.name, NULL);
		if (vi) {
			node->v.n_var->vr_inst = vi;
			vi->vi_flags |= SLV_VIF_USED;
		}
		break;

	case node_uminus:
	case node_not:
		nodefixup(node->v.n_arg[0], vit, expr);
		node->n_min_eval = node->v.n_arg[0]->n_min_eval;
		break;
		
	case node_add:
	case node_sub:
	case node_mul:
	case node_div:
	case node_or:
	case node_and:
	case node_eq:
	case node_ne:
 	case node_lt:
	case node_le:
	case node_gt:
	case node_ge:
	case node_pow:
		nodefixup(node->v.n_arg[0], vit, expr);
		nodefixup(node->v.n_arg[1], vit, expr);
		node->n_min_eval = MAX(node->v.n_arg[0]->n_min_eval,
				       node->v.n_arg[1]->n_min_eval);
		break;
		
 	case node_cond:
		nodefixup(node->v.n_arg[0], vit, expr);
		min_eval = node->v.n_arg[0]->n_min_eval;
		if (node->v.n_arg[1]) {
			nodefixup(node->v.n_arg[1], vit, expr);
			node->n_min_eval = MAX(min_eval,
					       node->v.n_arg[1]->n_min_eval);
		}
		nodefixup(node->v.n_arg[2], vit, expr);
		node->n_min_eval = MAX(min_eval,
				       node->v.n_arg[2]->n_min_eval);
		break;

	case node_func:
		min_eval = 0;
		for (np = node->v.n_func.args; np; np = np->n_next) {
			nodefixup(np, vit, expr);
			if (np->n_min_eval > min_eval)
				min_eval = np->n_min_eval;
		}
		node->n_min_eval += min_eval;
		break;
		
	case node_const:
		break;

	case node_subexpr: {
		struct slb_expression *subexpr =
			expression_lookupz(node->v.n_expr.name);
		if (!expr) {
			grecs_error(&subexpr->ex_locus, 0,
				    "unknown expression %s",
				    node->v.n_expr.name);
			return;
		}
		if (!node->v.n_expr.expr) {
			node->v.n_expr.expr = copy_expression(subexpr,
							      expr->ex_vartab);
			fixup_expression(node->v.n_expr.expr, vit);
			node->n_min_eval = node->v.n_expr.expr->ex_tree->n_min_eval;
		}
		break;
	}
		
	}
}

struct report_undef_closure
{
	grecs_locus_t *loc;
	int rc;
};

static int
report_undefined_refs(void *sym, void *data)
{
	struct slb_varref *ref = sym;
	struct report_undef_closure *clos = data;
	
	if (!ref->vr_inst) {
		grecs_error(clos->loc, 0,
			    _("variable or constant %s not defined"),
			    ref->vr_sym.name);
		clos->rc = 1;
	}
	return 0;
}

static int
mark_used_instances(void *sym, void *data)
{
	struct slb_varinstance *vinst = sym;
	grecs_locus_t *loc = data;

	bind_output_refname(vinst);
	if (!(vinst->vi_flags & SLV_VIF_USED)) {
		grecs_warning(loc, 0,
			      _("variable %s is never used"),
			      vinst->vi_sym.name);
	}
	return 0;
}

void
bind_expression(struct slb_expression *expr, struct symtab *vit)
{
	nodefixup(expr->ex_tree, vit, expr);
}

int
fixup_expression(struct slb_expression *expr, struct symtab *vit)
{
	struct report_undef_closure clos;

        if (expr->ex_fixed)
                return 0;
	nodefixup(expr->ex_tree, vit, expr);
	clos.rc = 0;
	clos.loc = &expr->ex_locus;
	symtab_enumerate(expr->ex_vartab, report_undefined_refs, &clos);
	undo_output_refs();
	symtab_enumerate(vit, mark_used_instances, &expr->ex_locus);
	return clos.rc;
}
